//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nwav;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8}
//==============================================================================
interface
//==============================================================================
uses
  nmaths,
  ntypes;

const
  maxchannels      = 8;
  maxblocksize     = 4096;

  CUEChunkMaxSize  = 65536;
  FACTChunkMaxSize = 65536;
  FMTChunkMaxSize  = 18+22;

  bitstocalculate  = 32;
  bytestocalculate = 4;

  buffsize         = maxblocksize*maxchannels*bytestocalculate*2;

  channel_offset : array[0..maxchannels-1] of integer = (0,integer_size*1*maxblocksize,integer_size*2*maxblocksize,integer_size*3*maxblocksize,integer_size*4*maxblocksize,integer_size*5*maxblocksize,integer_size*6*maxblocksize,integer_size*7*maxblocksize);

Type
  allChannelsSampleBlock = array[0..maxchannels-1, 0..maxblocksize-1] of Integer;
  allChannelsSampleBlockPtr = ^AllChannelsSampleBlock;

  tChunkID = array[0..3] of Char;

  tChunkHeader = packed record
    Case integer of
     0 : (ID:tChunkID; Size : Cardinal);
     1 : (CNum : Cardinal);
    End;

  tChunkMapRecord = Packed Record
    Header : tChunkHeader;
    CPtr   : Integer;
    Packer : Integer;
  End;

  tWAVHeader = Packed Record
    Header   : tChunkHeader; //"RIFF"
    RIFFType : tChunkID; //"WAVE"
    Packer   : Integer;
  end;

  tFMTChunk = Packed Record
    Header          : tChunkHeader; //"fmt "
    wFormatTag      : Word;
    wChannels       : Word;
    nSamplesPerSec  : Cardinal;
    nAvgBytesPerSec : Cardinal;
    nBlockAlign     : Word;
    wBitsPerSample  : Word;
    cbSize          : Word;
    Samples : Packed Record
    Case integer of
      0:(wValidBitsPerSample   : Word);
      1:(wSamplesPerBlock      : Word);
      2:(wReserved             : Word);
    End;
    dwChannelMask   : Cardinal;
    SubFormat_Word  : Array[0..3] of Word;
    SubFormat_Byte  : Array[0..7] of Byte;
  End;

  tFACTChunk = Packed Record
    Header : tChunkHeader; //"fact"
    Data : array[0..65535] of char;
  End;

  tDATAChunk = Packed Record
    Header : tChunkHeader; //"data"
  End;

  tlossyWAVChunk = Packed Record
   WAV  : tWAVHeader;
   FMT  : tFMTChunk;
   FACT : tFACTChunk;
   DATA : tDataChunk;
  End;

  tWaveBuffer = array[0..Buffsize-1] of byte;

var
  btrd_codec_block : AllChannelsSampleBlock;
  corr_codec_block : AllChannelsSampleBlock;

  rotating_blocks     : array[0..3] of AllChannelsSampleBlock;
  rotating_blocks_ptr : array[0..3] of AllChannelsSampleBlockPtr = (@rotating_blocks[0],@rotating_blocks[1],@rotating_blocks[2],@rotating_blocks[3]);

  prev_codec_block_size  : Integer = 0;
  last_codec_block_size  : Integer = 0;
  this_codec_block_size  : Integer = 0;
  next_codec_block_size  : Integer = 0;

  this_codec_block_size_recip : Double = 0;

function channelCount: integer;                             // number of channels
function sampleRate: Integer;                               // sample rate in Hz
function bitsPerSample: integer;                            // bits per sample
function bytesPerSample: integer;                           // bytes per sample
function totalsamples : integer;                            // total samples in WAV file.

procedure MergeFiles;                                       // Merge lossy.wav and lwcdf.wav files.

procedure openWavIO(var blocksize: Integer);                // Zero if operation not succesful.
                                                            // Reads the header chunks of inWav and memorizes the
                                                            // Headerrmation from the fmt chunk.
                                                            // Creates outWav and fills its header chunks with an
                                                            // identical copy of that of inWav.
                                                            // Memorizes blocksize (nr of samples to read/write with subsequent calls to
                                                            // readNextSampleBlock resp. writeNextSampleBlock).

function readNextNextCodecBlock : integer;                  // False if not possible (usually because all the samples have been read).
                                                            // sampleBlock: next block of samples from input wav file for all the channels.
                                                            //              first sample is in element with index 0.
                                                            // readsamplesCount: number of samples read (in each of the channels).

function writeNextbtrdcodecblock : integer;                 // False if not possible.
                                                            // Writes sampleBlock as the next sample block of all the channels to the output file.
                                                            // The number of samples to be written is given by samplesCountToBeWritten.
                                                            // First sample written is sampleBlock element with index overlapsize
                                                            // - overlapsize from openWavIO call.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.

function writeNextcorrcodecblock : integer;                 // False if not possible.
                                                            // Writes sampleBlock as the next sample block of all the channels to the output file.
                                                            // The number of samples to be written is given by samplesCountToBeWritten.
                                                            // First sample written is sampleBlock element with index overlapsize
                                                            // - overlapsize from openWavIO call.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.

function closeWavIO : integer;                              // False if not possible.
                                                            // Finalizes the wavIO operations.

function datestamp_string : String;                         // Process version, date & time.

//==============================================================================
implementation
//==============================================================================
{$I-}
uses
  Math,
  Windows,
  SysUtils,
  nparameter;

//==============================================================================
//nwav constants, variables & types
//==============================================================================
Var
  WAVBuff                             : tWaveBuffer;
  BtrdBuff                            : tWaveBuffer;
  Corrbuff                            : tWaveBuffer;
  MergeBuff                           : tWaveBuffer;

  chunkspace                          : array[0..2097151] of byte;
  chunkmap                            : array[0..1023] of tChunkMapRecord;

  WAVEFile, Btrdfile, Corrfile        : File;

  datetimestring                      : string;
  datestamp                           : String;

  ChunkHeader                         : tChunkHeader;
  WAVEChunk                           : tlossyWAVChunk;
  BTRDChunk                           : tlossyWAVChunk;
  CORRChunk                           : tlossyWAVChunk;

  filesOpened                         : Cardinal;
  readresult                          : Cardinal;
  writeresult                         : Cardinal;

  buffsizeread                        : Cardinal; // number of byte that make up for inbuff
  buffsizewrite                       : Cardinal; // number of byte that make up for outbuff
  nrOfByteInOneBlockInBuff            : Cardinal; // number of byte that make up a complete block in inBuff and outBuff.
  nrOfBlocksInReadBuff                : Cardinal;
  nrOfBlocksInWriteBuff               : Cardinal;

  nrOfFullBlocksReadIntoInBuff        : Cardinal; // with the last BlockRead: how many full blocks managed it into inBuff ?
  nrOfByteOfTheLastIncompleteBlock    : Cardinal; // when EOF was encountered with the last BlockRead: the number of byte that didn't make it
                                                  // into the last full block
  nrOfBlockInBuffNotYetFetched        : Cardinal; // the next number of the block in inBuff that was not fetched yet by readNextSampleBlock.
                                                  // 0: no data in inBuff;
  nrOfByteFilledInBtrdBuff            : Cardinal; // how many byte have managed it into outBuff that are not written to the outputfile yet?
  nrOfByteFilledInCorrBuff            : Cardinal;  // how many byte have managed it into outBuff that are not written to the outputfile yet?
  bitdepth                            : Cardinal;
  bytedepth                           : Cardinal;

  filereadmode                        : Integer;
  sampleByteLeftToRead                : Cardinal;
  samplebytesread                     : Cardinal;
  readbyteCount                       : Cardinal;
  byteslefttoread                     : cardinal;
  thisreadbyteCount                   : Cardinal;

  writebytecount                      : Cardinal;
  thiswritebytecount                  : Cardinal;
  firstsamplesread                    : Cardinal;
  lastwinerror                        : Cardinal;
  lastIOResult                        : Cardinal;
  datachunkpos                        : Cardinal;
  ThisFilePos                         : Cardinal;

  retries                             : Cardinal;
  CantReadInWavAnyMore                : Cardinal;

  numchunks                           : Cardinal = 0;
  endofchunkmap                       : Cardinal;

  this_scale                          : Double;
  this_scale_inv                      : Double;
  buffsizetouse                       : Cardinal;
  totalsamples_int                    : Cardinal;
  last_find_text                      : integer = -1;

  nwav_packer                         : array[0..2] of integer;

//==============================================================================

  function channelCount: integer;
  begin
    channelCount:=WAVEChunk.FMT.wChannels;
  end;

  function sampleRate: Integer;
  begin
    sampleRate:=WAVEChunk.FMT.nSamplesPerSec;
  end;

  function bitsPerSample: integer;
  begin
    bitsPerSample:=bitdepth;
  end;

  function bytesPerSample: integer;
  begin
    bytesPerSample:=bytedepth;
  end;

  function totalsamples : integer;
  begin
    totalsamples:=totalsamples_int;
  end;

//==============================================================================

  procedure combinelossybuffers(bytes_in_buffer : Integer);
  var
    this_sample, btrd_sample, corr_sample : integer;
  begin
    case bytedepth of
      1 : lossyWAVerror('Invalid bitdepth.',$12);

      2 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,2
                cmp    ecx,0
                jl     @@01

                movsx  eax,word ptr [ebx+ecx]
                mov    dword ptr btrd_sample,eax

                movsx  eax,word ptr [edi+ecx]
                mov    dword ptr corr_sample,eax

                fild   btrd_sample
                fmul   this_scale_inv
                fiadd  corr_sample
                fistp  this_sample

      @@00b:    mov    eax,this_sample
                mov    word ptr [esi+ecx],ax
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
          end;

      3 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,3
                cmp    ecx,0
                jl     @@01

                xor    eax,eax
                movsx  ax,byte ptr [ebx+ecx+2]
                shl    eax,16
                mov    ax,word ptr [ebx+ecx]
                mov    dword ptr btrd_sample,eax

                xor    eax,eax
                movsx  ax,byte ptr [edi+ecx+2]
                shl    eax,16
                mov    ax,word ptr [edi+ecx]
                mov    dword ptr corr_sample,eax

                fild   btrd_sample
                fmul   this_scale_inv
                fiadd  corr_sample
                fistp  this_sample

                mov    eax,this_sample
                mov    word ptr [esi+ecx],ax
                shr    eax,16
                mov    byte ptr [esi+ecx+2],al
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
          end;

      4 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,4
                cmp    ecx,0
                jl     @@01

                fild   dword ptr [ebx+ecx]
                fmul   this_scale_inv
                fiadd  dword ptr [edi+ecx]
                fistp  dword ptr [esi+ecx]
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
         end;
    end;
  end;

//==============================================================================

  procedure openfile(var thisfile : file; thisname : string; thisfilemode : cardinal; orvalue : cardinal);
  begin
    filemode:=thisfilemode;

    AssignFile(thisfile, thisname);
    if IOResult <> 0 then exit;

    case thisfilemode of
      0 : Reset(thisfile, 1);
      2 : Rewrite(thisfile,1);
    else
      lossyWAVerror('Invalid FileMode!',$12);
    end;
    if IOResult <> 0 then exit;

    filesOpened:=filesopened or orvalue;
  end;

//==============================================================================

  procedure wavIOExitProc(wavIOString : String; wavIOCode : Integer);
  var
    wi : integer;
  begin
    for wi:=0 to numchunks-1 do
      write(ErrOutput,chunkmap[wi].Header.ID,',');
    write(ErrOutput,crlf);
    write(ErrOutput,wavIOString,crlf);
    closeWavIO;
    exitcode:=wavIOcode;
    halt;
  end;

//==============================================================================

  function readfromWAVEfile(buffpointer : pointer; bytestoread : Cardinal) : cardinal;
  begin
    readbytecount:=0;
    retries:=0;

    case filereadmode of
      // Normal file read;
      0 : begin
            blockread(wavefile,buffpointer^,bytestoread,readbytecount);
            lastIOResult:=IOResult;
            if (lastIOResult<>0) or (readbytecount<bytestoread) then
              cantreadinwavanymore:=1;
          end;

      // Read from pipe and use Windows API GetLastError to determine whether pipe is broken;
      2 : while (readbytecount<bytestoread) and (CantReadInWavAnyMore=0) do
          begin
            byteslefttoread:=bytestoread-readbytecount;

            if ReadFile(hSTDIN, pointer(cardinal(buffpointer)+readbytecount)^, byteslefttoread, thisreadbytecount,nil) then
              lastwinerror:=0
            else
              lastwinerror:=getlasterror();

            if thisreadbytecount=0 then
            begin
              retries:=retries+1;
              ndelay(powersoftwoinv[11-min(7,retries)]);
            end
            else
              retries:=0;

            if (lastwinerror=error_broken_pipe) or (retries=32) then
              cantreadinwavanymore:=1;

            readbytecount:=readbytecount+thisreadbytecount;
          end;
    end;

    readfromWAVEfile:=readbytecount;
  end;

//==============================================================================

  function WordAlign(WAvalue : Cardinal) : Cardinal;
  begin
    WordAlign:=WAvalue + (WAvalue and 1);
  end;

//==============================================================================

  procedure WriteChunks(var outputfile : file);
  var
    thischunk : Cardinal;
  begin
    if ((merging=1) or (stdoutput=1)) and (WAVEChunk.FACT.Header.Size<>0) then
      WAVEChunk.WAV.Header.Size:=WAVEChunk.WAV.Header.Size-(WAVEChunk.FACT.Header.Size+8);

    BlockWrite(outputfile, WAVEChunk.WAV, 12);
    if IOResult <> 0 then exit;

    for thischunk:=0 to numchunks-1 do
    begin
      if chunkmap[thischunk].cptr=-1 then
      begin
        case chunkmap[thischunk].header.cnum of
          $20746D66 : begin //'fmt '
                        BlockWrite(outputfile, WAVEChunk.FMT, WordAlign(WAVEChunk.FMT.Header.Size+8));
                        if IOResult <> 0 then
                          wavIOExitProc('Error writing ''fmt '' chunk.',$12);
                        if (merging=0) and (WaveChunk.Fact.Header.Size<>0) and (stdoutput=0) then
                        begin
                          BlockWrite(outputfile, WAVEChunk.FACT, WordAlign(WAVEChunk.FACT.Header.Size+8));
                          if IOResult <> 0 then
                            wavIOExitProc('Error writing ''fact'' chunk.',$12);
                        end;
                      end;
          $74636166 : begin //'fact'
                      end;
        end
      End
      else
      Begin
        Blockwrite(outputfile,chunkmap[thischunk].header,8);
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
        Blockwrite(outputfile,chunkspace[chunkmap[thischunk].cptr],WordAlign(chunkmap[thischunk].header.size));
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
      end;
    end;

   if stdoutput=0 then
     datachunkpos:=filepos(outputfile);
    BlockWrite(outputfile, WAVEChunk.DATA, 8);
    if IOResult <> 0 then exit;
  end;

  procedure RewriteDataChunk(var outputfile : file);
  begin
    thisfilepos:=filepos(outputfile);
    seek(outputfile,0);
    wavechunk.WAV.header.size:=thisfilepos-8;
    blockwrite(outputfile,wavechunk.WAV,8,writeresult);

    seek(outputfile,datachunkpos);
    wavechunk.data.header.size:=samplebytesread;
    blockwrite(outputfile,wavechunk.data,8,writeresult);
  end;
//==============================================================================

  function HexChar(HC : Integer) : Char;
  Asm
                Mov    eax,hc
                and    eax,15
                cmp    eax,9
                jle    @@01
                add    eax,7
      @@01:     add    eax,48
  End;

  function Dec2Hex(d2h : integer) : String;
  Begin
    Dec2Hex:=HexChar(d2h shr 28)+HexChar(d2h shr 24)+HexChar(d2h shr 20)+HexChar(d2h shr 16)+HexChar(d2h shr 12)+HexChar(d2h shr 8)+HexChar(d2h shr 4)+HexChar(d2h)
  End;

//==============================================================================

  procedure ReadChunks(var ThisFile : File; var ThisChunk : tlossyWAVChunk);
  begin
    fillchar(thischunk,sizeof(thischunk),0);
    numchunks:=0;
    endofchunkmap:=0;

    BlockRead(ThisFile,ChunkHeader,8,readresult);
    if readresult<>8 then
      wavIOExitProc('Error Reading WAV file Information.',$11);

    if chunkheader.id='RIFF' then
    begin
      move(chunkheader,thischunk.wav,8);
      blockread(thisfile,thischunk.wav.rifftype,4,readresult);
      if (readresult<>4) then
        wavIOExitProc('Error Reading WAV file Information.',$11);
    end
    else
      wavIOexitProc('File type is not RIFF.',$12);

    if ThisChunk.WAV.RIFFType <> 'WAVE' then
      wavIOExitProc('RIFF type is not WAVE.',$12);

    repeat
      BlockRead(ThisFile,ChunkHeader,8,readresult);
      if readresult<>8 then
        wavIOExitProc('Error Reading WAV file Information.',$11);

      case chunkheader.cnum of
        $20746D66 : begin //'fmt '
                      chunkmap[numchunks].Header:=chunkheader;
                      chunkmap[numchunks].Cptr:=-1;
                      numchunks:=numchunks+1;
                      move(chunkheader,ThisChunk.FMT,8);
                      Case ThisChunk.FMT.Header.Size of
                        16,18,40 : Begin
                                   End;
                      else
                        WavIOExitProc('FMT Chunk incorrect size.',$12);
                      End;
                      blockread(ThisFile,ThisChunk.FMT.wFormatTag,WordAlign(thischunk.FMT.Header.Size),readresult);
                      if readresult<>wordalign(thischunk.FMT.Header.Size) then
                        wavIOExitProc('Error reading FMT chunk.',$12);
                    end;
        $61746164 : begin //'data'
                      Move(ChunkHeader,ThisChunk.DATA,8);
                    end;
      else
        chunkmap[numchunks].Header:=chunkheader;
        chunkmap[numchunks].Cptr:=EndofChunkMap;
        If endofchunkmap>sizeof(chunkspace) then
          wavIOExitProc('Too much WAV information before ''data'' chunk!',$12);
        blockread(thisfile,chunkspace[endofchunkmap],WordAlign(chunkheader.size),readresult);
        if readresult<>wordalign(chunkheader.size) then
          wavIOExitProc('Error Reading '''+chunkheader.ID+''' Chunk.',$12);
        endofchunkmap:=endofchunkmap+wordalign(chunkheader.size);

        if (chunkmap[numchunks].Header.Id='fact') then
        begin
          ThisChunk.FACT.Header:=ChunkMap[numchunks].Header;
          if thischunk.fact.header.size<factchunkMaxsize then
            move(chunkspace[chunkmap[numchunks].Cptr],thischunk.fact.data,thischunk.fact.header.size);
          if copy(thischunk.fact.data,1,8)<>'lossyWAV' then
            thischunk.fact.header.size:=0
          else
            chunkmap[numchunks].cptr:=-1;
        end;

        numchunks:=numchunks+1;
      end;
    until ChunkHeader.ID='data';

    if (ThisChunk.FMT.wFormatTag<>1) and ((ThisChunk.FMT.wformatTag=$FFFE) and (ThisChunk.FMT.SubFormat_Word[0]<>1)) then
      wavIOExitProc('WAV Data is not uncompressed PCM.',$12);
  end;

//==============================================================================

  function find_text(find_str : string; source_str : string) : integer;
  var
    ft_i : Integer;
    ft_l : Integer;
  begin
    ft_i:=0;
    ft_l:=length(find_str);

    if ft_l=0 then
      ft_i:=-1
    else
    begin
      repeat
        inc(ft_i);
      until (ft_i=(length(source_str)-ft_l)) or (find_str=copy(source_str,ft_i,ft_l));

      if ft_i=(length(source_str)-ft_l) then
        ft_i:=-1
    end;
    find_text:=ft_i;
    last_find_text:=ft_i;
  end;

//==============================================================================

  procedure MergeFiles;
  var
    thisread   : cardinal;
    numreads   : cardinal;
    lastread   : cardinal;
    readcount  : cardinal;
    this_error : Integer;
    temp_str   : String;
  begin
    filesOpened:=0;

    openfile(BTRDFile,LossyInp,0,2);
    ReadChunks(BTRDFile,BTRDChunk);

    openfile(CORRFile,LwcdfInp,0,2);
    ReadChunks(CORRFile,CORRChunk);

    write(ErrOutput,crlf,'Merging: ',lossyfilename,crlf);
    write(ErrOutput,     '       & ',lwcdffilename,crlf);

    if btrdchunk.wav.Header.Size<>corrchunk.wav.Header.Size then
      lossyWAVerror('Files not same length.',$12);

    if btrdchunk.data.Header.Size<>corrchunk.data.Header.Size then
      lossyWAVerror('Data not same length.',$12);

    if (btrdchunk.fact.header.size=0) then
      lossyWAVerror('FACT Data missing (lossy.wav).',$12);

    if (corrchunk.fact.header.size=0) then
      lossyWAVerror('FACT Data missing (lwcdf.wav).',$12);

    if btrdchunk.fact.data<>corrchunk.fact.data then
      lossyWAVerror('FACT Data mismatch.',$12);

    if copy(btrdchunk.fact.data,1,8)<>'lossyWAV' then
      lossyWAVerror('Not a lossyWAV FACT chunk.',$12);

    if (btrdchunk.fmt.Header.Size<>corrchunk.fmt.Header.Size) or
       (btrdchunk.fmt.wFormatTag<>corrchunk.fmt.wFormatTag) or
       (btrdchunk.fmt.wChannels<>corrchunk.fmt.wChannels) or
       (btrdchunk.fmt.nSamplesPerSec<>corrchunk.fmt.nSamplesPerSec) or
       (btrdchunk.fmt.nAvgBytesPerSec<>corrchunk.fmt.nAvgBytesPerSec) or
       (btrdchunk.fmt.nblockAlign<>corrchunk.fmt.nBlockAlign) or
       (btrdchunk.fmt.wBitsPerSample<>corrchunk.fmt.wBitsPerSample) then
      lossyWAVerror('FMT Data mismatch.',$12);

    bytedepth:=(btrdchunk.fmt.wBitsPerSample+7) shr 3;
    if bytedepth<2 then
      lossyWAVerror('Invalid bitdepth: '+inttostr(btrdchunk.fmt.wBitsPerSample),$12);

    lastread:=btrdChunk.Data.Header.Size mod buffsize;
    numreads:=btrdchunk.data.Header.Size div buffsize+cardinal(lastread>0);

    move(btrdchunk,wavechunk,sizeof(btrdchunk));
    openfile(WAVEFile,WAVOut,2,1);
    writechunks(WAVEfile);

    if find_text('--scale ',wavechunk.fact.data)<>-1 then
      temp_str:=copy(wavechunk.fact.data,last_find_text+7,8)
    else
      temp_str:='1.000000';

    val(temp_str,this_scale,this_error);
    if this_error<>0 then
      lossyWAVerror('Invalid scale value: '+temp_str,$12);

    this_scale_inv:=1/this_scale;

    buffsizetouse:=buffsize-(buffsize mod (bytedepth*btrdchunk.fmt.wChannels));

    for readcount:=0 to numreads-1 do
    begin
      thisread:=min(buffsizetouse,btrdchunk.data.Header.Size-Readcount*buffsizetouse);
      blockread(BTRDFile,BTRDBuff,thisread,readresult);
      blockread(CORRFile,CORRBuff,thisread,readresult);
      readbytecount:=readbytecount+readresult;
      combinelossybuffers(thisread);
      blockwrite(WAVEFile,MERGEBuff,thisread,readresult);
      write(ErrOutput,#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    close(CORRFile);
    readbytecount:=filepos(btrdfile);
    readresult:=65536;

    while readresult=65536 do
    begin
      blockread(BTRDFile,BTRDBuff,65536,readresult);
      blockwrite(WAVEFile,btrdbuff,readresult);
      readbytecount:=readbytecount+readresult;
      write(ErrOutput,#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    write(ErrOutput,crlf);

    close(BTRDFile);
    close(WAVEFile);
  end;

//==============================================================================
  procedure openWavIO(var blocksize: Integer);
  begin
    filesOpened:=0;
    filereadmode:=(STDINPUT*2);
    blocksize:=0;

    openfile(WAVEFile,wavinp,0,1);
    ReadChunks(WAVEFile,WAVEChunk);

    if wavechunk.data.header.size=0 then
      wavIOexitproc('Zero sample WAV file!?!',$12);

    bitdepth:=WAVEChunk.FMT.wBitsPerSample;
    if (bitdepth<9) or (bitdepth>32) then
    begin
      write(ErrOutput,'Input file has a bitdepth of ' + IntToStr(bitdepth) + ' bits per sample.'+crlf);
      wavIOExitProc('This is not supported.',$12);
    end;
    bytedepth:=(bitdepth+7) shr 3;

    totalsamples_int:=floor(1.0*WAVEChunk.DATA.Header.Size / WAVEChunk.FMT.wChannels / bytedepth);

    if WAVEChunk.FMT.nSamplesPerSec<32000 then
    begin
      write(ErrOutput,'Sample Rate too low : '+floattostrf(samplerate/1000,ffFixed,99,1)+'Hz (Min=32kHz).',crlf);
      wavIOExitProc('This is not supported.',$12);
    end;

    if WAVEChunk.FMT.nSamplesPerSec>384000 then
    begin
      write(ErrOutput,'Sample Rate too high : '+floattostrf(samplerate/1000,ffFixed,99,1)+'Hz (Max=384kHz).',crlf);
      wavIOExitProc('This is not supported.',$12);
    end;

    case WAVEChunk.FMT.nSamplesPerSec of
       32000.. 69075 : blocksize:= 512;
       69076..138150 : blocksize:=1024;
      138151..276300 : blocksize:=2048;
    else
      blocksize:=4096;
    end;
  
    sampleByteLeftToRead:=WAVEChunk.DATA.Header.Size;
    nrOfByteInOneBlockInBuff:=bytedepth * cardinal(WAVEChunk.FMT.wChannels * blocksize);

    nrOfBlocksInReadBuff:=(1+(buffsize div nrOfByteInOneBlockInBuff-1)*cardinal(stdinput=0));
    nrOfBlocksInWriteBuff:=(1+(buffsize div nrOfByteInOneBlockInBuff-1)*cardinal(stdoutput=0));

    buffsizeread :=nrOfByteInOneBlockInBuff * nrOfBlocksInReadBuff;
    buffsizewrite:=nrOfByteInOneBlockInBuff * nrOfBlocksInWriteBuff;

    CantReadInWavAnyMore:=0;
    //==========================================================================

    if checking=1 then
    begin
      if (WAVEChunk.FACT.Header.Size>0) and (copy(WAVEChunk.FACT.Data,1,8)='lossyWAV') then
      begin
        exitcode:=$10;
        if (verbosity=1) and (silent=0) then
        begin
          write(ErrOutput,'lossyWAV FACT Chunk found. File already processed.',crlf);
          write(ErrOutput,WAVEChunk.FACT.Data);
        end;
      end
      else
      begin
        exitcode:=$00;
        if (verbosity=1) and (silent=0) then
          write(ErrOutput,'lossyWAV FACT Chunk not found. File not marked as processed.',crlf);
      end;
      closeWAVio;
      halt;
    end
    else
    begin
      if copy(WAVEChunk.FACT.data,1,8)='lossyWAV' then
      begin
        write(ErrOutput,WAVEChunk.FACT.Data);
        lossyWAVerror('lossyWAV FACT Chunk found. File already processed.',$10);
      end;
    end;

    //==========================================================================
    //create 'fact' chunk data string.
    //==========================================================================
    datetimetostring(datestamp,'',now);
    WAVEChunk.FACT.Header.ID:='fact';
    DateTimeString:=version_string+' @ '+datestamp+', '+parameter_string+crlf+#0;
    if length(DateTimestring) and 1=1 then
      DateTimestring:=DateTimestring+#0;
    move(DateTimestring[1],WAVEChunk.FACT.data,length(DateTimestring));
    WAVEChunk.FACT.Header.Size:=length(DateTimestring);
    inc(WAVEChunk.WAV.Header.Size,WAVEChunk.FACT.Header.Size+8);
    //==========================================================================

    firstsamplesread:=min(samplebytelefttoread,buffsizeread);
    if (readfromWAVEfile(@WAVBuff, firstsamplesread)<firstsamplesread) or (cantreadinwavanymore=1) then
    begin
      CantReadInWavAnyMore:=1;
      exit;
    end;

    samplebytesread:=readbytecount;
    sampleByteLeftToRead:=sampleByteLeftToRead - readByteCount;
    nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
    nrOfByteOfTheLastIncompleteBlock:=readbyteCount - nrOfFullBlocksReadIntoInBuff * nrOfByteInOneBlockInBuff;
    nrOfBlockInBuffNotYetFetched:=1;
    nrOfByteFilledInBtrdBuff:=0;
    nrOfByteFilledInCorrBuff:=0;

    //==========================================================================

    openfile(BtrdFile,lossyout,2,2);
    WriteChunks(btrdfile);

    if (STDOUTPUT=0) and (correct=1) then
    begin
      openfile(CorrFile,lwcdfout,2,4);
      WriteChunks(Corrfile);
    end;

    //==========================================================================
  end;

//==============================================================================

  procedure read_transfer(pStartB: PByte; pEndB: PByte);
  begin
    case bytedepth of
      2 : asm
                push   ebx
                push   edi

                xor    edi,edi
                mov    ecx,pStartB

      @@20:     mov    ebx,0

      @@21:     movsx  eax,word ptr [ecx]
                add    ecx,2

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.wChannels
                jl     @@21

                add    edi,1
                cmp    ecx,pEndB
                jl     @@20

                pop    edi
                pop    ebx
          end;

      3 : asm
                push   ebx
                push   edi

                xor    edi,edi
                mov    ecx,pStartB

      @@30:     mov    ebx,0

      @@31:     movsx  eax,word ptr [ecx+1]
                shl    eax,8
                mov    al,byte ptr [ecx]
                add    ecx,3

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.wChannels
                jl     @@31

                add    edi,1
                cmp    ecx,pEndB
                jl     @@30

                pop    edi
                pop    ebx
          end;

      4 : asm
                push   ebx
                push   edi

                xor    edi,edi
                mov    ecx,pStartB

      @@40:     mov    ebx,0

      @@41:     mov    eax,[ecx]
                add    ecx,4

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.wChannels
                jl     @@41

                add    edi,1
                cmp    ecx,pEndB
                jl     @@40

                pop    edi
                pop    ebx
          end;
    end;
  end;

//==============================================================================

  function readNextNextCodecBlock : integer;
  var
    pB, pStartB: PByte;
    thisblockreadlength : cardinal;
  begin
    readNextnextCodecBlock:=0;
    Next_Codec_Block_Size:=0;

    if cantreadinwavanymore=1 then
      exit;

    if nrOfBlockInBuffNotYetFetched = 0 then // samples to be fetched aren't in inbuff
    begin
      if (nrOfFullBlocksReadIntoInBuff < nrOfBlocksInReadBuff) or (sampleByteLeftToRead=0) then // EOF encountered with last call to readNextSampleBlock
        exit; // no more data

      thisblockreadlength:=min(buffsizeread,samplebytelefttoread);
      readbytecount:=0;
      retries:=0;

      readfromWAVEfile(@WAVBuff,thisblockreadlength);

      sampleByteLeftToRead:=sampleByteLeftToRead-readByteCount;
      samplebytesread:=samplebytesread+readbytecount;

      nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
      nrOfByteOfTheLastIncompleteBlock:=integer(readbyteCount) mod - nrOfByteInOneBlockInBuff;

      nrOfBlockInBuffNotYetFetched:=1;
    end;

    pB:=@WAVBuff;
    Inc(pB, (nrOfBlockInBuffNotYetFetched - 1) * nrOfByteInOneBlockInBuff);
    pStartB:=pB; // points to first sample in Block

    if nrOfBlockInBuffNotYetFetched <= nrOfFullBlocksReadIntoInBuff then
    begin
      Inc(pB, nrOfByteInOneBlockInBuff);  // pB points to the spot after last sample in Block
      next_codec_block_size:=nrOfByteInOneBlockInBuff DIV (bytedepth * WAVEChunk.FMT.wChannels);
      nrOfBlockInBuffNotYetFetched:=nrOfBlockInBuffNotYetFetched + 1;
      if nrOfBlockInBuffNotYetFetched > nrOfBlocksInReadBuff then
        nrOfBlockInBuffNotYetFetched:=0;
    end
    else
    begin
      if nrOfByteOfTheLastIncompleteBlock = 0 then
        exit; // no more data
      Inc(pB, nrOfByteOfTheLastIncompleteBlock);  // pB points to the spot after last sample in Block
      next_codec_block_size:=nrOfByteOfTheLastIncompleteBlock DIV (bytedepth * WAVEChunk.FMT.wChannels);
      nrOfBlockInBuffNotYetFetched:=0; // force a try for a new Blockread with next call which will lead to a return False
    end;

    read_transfer(pStartB, pB);
    readNextNextCodecBlock:=1;
  end;

//==============================================================================
  Procedure WriteTransfer(var OutputCodecBlock : AllChannelsSampleBlock; pb : pByte);
  Begin
    case bytedepth of
      2 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    word ptr [ecx],ax
                add    ecx,2

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;

      3 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    word ptr [ecx],ax
                shr    eax,16
                mov    byte ptr [ecx+2],al
                add    ecx,3

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;

      4 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    [ecx],eax
                add    ecx,4

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;
    end;
  End;

  function writeNextCodecBlock(var outputfile : file; var OutBuff; var OutputCodecBlock : AllChannelsSampleBlock; var FilledOutputBuffer : Cardinal): Cardinal;
  var
    nrOfByteForOutBuff : Cardinal;
    pB: PByte;
  begin
    writeNextCodecBlock:=0;
    nrOfByteForOutBuff:=integer(bytedepth) * WAVEChunk.FMT.wChannels * this_codec_block_size;
    writebytecount:=0;

    if FilledOutputBuffer + nrOfByteForOutBuff > buffsizewrite then // not sufficient space in outBuff : flush outBuff
    begin
      BlockWrite(outputfile, OutBuff, FilledOutputBuffer);
      if IOResult <> 0 then
        exit;
      FilledOutputBuffer:=0;
    end;

    pB:=@OutBuff;
    Inc(pB, FilledOutputBuffer); // points to first free spot in outBuff
    WriteTransfer(OutputCodecBlock,pb);
    FilledOutputBuffer:=FilledOutputBuffer + nrOfByteForOutBuff;
    writeNextCodecBlock:=1;
  end;

  function writeNextBtrdCodecBlock: integer;
  begin
    WriteNextBtrdCodecBlock:=WriteNextCodecBlock(BtrdFile,BtrdBuff,btrd_codec_block,nrofByteFilledInBtrdBuff);
  end;

  function writeNextCorrCodecBlock : integer;
  begin
    WriteNextCorrCodecBlock:=WriteNextCodecBlock(CorrFile,Corrbuff,corr_codec_block,nrofByteFilledInCorrBuff);
  end;

//==============================================================================

  function closeWavIO: integer;
  begin
    closeWavIO:=0;

    if filesOpened>1 then
    begin
      if (filesopened and 2=2) and (nrOfByteFilledInBtrdBuff > 0) then
      begin
        BlockWrite(BtrdFile, BtrdBuff, nrOfByteFilledInBtrdBuff);
        if filepos(Btrdfile) and 1=1 then
          BlockWrite(BtrdFile,nWAV_Packer[0],1);
        if IOResult <> 0 then exit;
      end;

      if (filesopened and 4=4) and (nrOfByteFilledInCorrBuff > 0) then
      begin
        BlockWrite(CorrFile, CorrBuff, nrOfByteFilledInCorrBuff);
        if filepos(Corrfile) and 1=1 then
          BlockWrite(CorrFile,nWAV_Packer[0],1);
        if IOResult <> 0 then exit;
      end;

      While not(eof(wavefile)) and (CantReadInWAVAnyMore=0) do
      Begin
        if (readfromWAVEfile(@WAVBuff, buffsizeread)<cardinal(buffsizeread)) or (IOResult <> 0) then
          CantReadInWavAnyMore:=CantReadInWavAnyMore or 1;

        if readbytecount<>0 then
        Begin
          if (filesopened and 2=2) then
          begin
            BlockWrite(BtrdFile, WAVBuff, readbyteCount);
            if IOResult <> 0 then exit;
          end;

          if filesopened and 4=4 then
          Begin
            BlockWrite(CorrFile, WAVBuff, readbyteCount);
            if IOResult <> 0 then exit;
          End;
        End;
      End;

      closeFile(WAVEFile);
      if IOResult <> 0 then exit;
      filesopened:=filesopened xor 1;

      if filesopened and 2=2 then
      begin
        if filepos(Btrdfile) and 1=1 then
          BlockWrite(BtrdFile,nWAV_Packer[0],1);
        if (stdinput=1) and (stdoutput=0) then
          rewritedatachunk(BtrdFile);
        CloseFile(BtrdFile);
        if IOResult <> 0 then exit;
        filesopened:=filesopened xor 2;
      end;

      if filesopened and 4=4 then
      begin
        if filepos(Corrfile) and 1=1 then
          BlockWrite(CorrFile,nWAV_Packer[0],1);
        if (stdinput=1) and (stdoutput=0) then
          rewritedatachunk(CorrFile);
        closeFile(CorrFile);
        if ioresult<>0 then exit;
        filesopened:=filesopened xor 4;
      end;
    end;

    closeWavIO:=1;
  end;

  function datestamp_string : String;
  begin
    datestamp_string:=datestamp;
  end;
{$I+}
//==============================================================================
initialization
//==============================================================================
  nwav_packer[0]:=0;
//==============================================================================
end.
//==============================================================================

